/* Alarm clock with Teensy 3.1 and audio output
 * 
 * alarm_clock_1_2.ino
 * 
 * www.weigu.lu/b/alarmclock
 * 
 * Audio file names on SD Card: 1.WAV, 2.WAV etc.
 * 
 * Copyright (C) 2015 Guy WEILER www.weigu.lu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <Audio.h>
#include <Wire.h>
#include <SPI.h>
#include <SD.h>
#include <DCF77.h> //https://github.com/thijse/Arduino-Libraries/downloads
#include <Time.h>  //RTC Teensy 3
#include <EEPROM.h>
#include <Encoder.h>
#include "Adafruit_LEDBackpack.h"
#include "Adafruit_GFX.h"
#include "Adafruit_TPA2016.h"

enum State { CLOCK, ALARM, SET_ALARM, MUSIC, DCF_77};
State state = CLOCK;

// create objects for 2 encoder
Encoder knobLeft(0, 2);
Encoder knobRight(6, 4);
long positionLeft  = -999; // dummy value foe encoders
long positionRight = -999;
int newLeft, newRight;
unsigned int rotaryTimer = 0;
// 7seg. display
Adafruit_7segment disp = Adafruit_7segment();
unsigned int seg7time, alarmtime, alarmtimeHour, alarmtimeMin, brightness;
const unsigned int brightnessDay = 2;
const unsigned int brightnessNight = 0;
const unsigned int brightnessDayTime = 600;
const unsigned int brightnessNightTime = 2200;
const unsigned int DCF77Time = 400;
// audio amp
Adafruit_TPA2016 audioamp = Adafruit_TPA2016();
unsigned int musicGain;
// 2 pushbuttons
const unsigned int PB_RIGHT = 5;
const unsigned int PB_LEFT = 1;
volatile int mainEventFlags = 0; // 16 Flags: in one interger
const unsigned int FLAG_INTERRUPT_PB_RIGHT = 0x01; // right button
const unsigned int FLAG_INTERRUPT_PB_LEFT  = 0x02; // left  button
const unsigned int FLAG_ALARM_SET          = 0x04; // alarm set
const unsigned int FLAG_SET_SONG           = 0x08; // set song number
const unsigned int FLAG_SET_GAIN           = 0x10; // set gain in music state
const unsigned int FLAG_DCF77_ERROR        = 0x20; // if no DCF77 connection
const unsigned int FLAG_SET_BRIGHTNESS     = 0x40; // set display brightness
// connect dac to SD-Card
AudioPlaySdWav           playWav1; 
AudioOutputAnalog        dac1; 
AudioConnection          patchCord1(playWav1, 0, dac1, 0);
// files on SD card
char fileName[13];
unsigned int fileNr = 1;
unsigned int filesondisk;
// DCF77
const unsigned int DCF_PIN = 23;
const unsigned int DCF_PIN_PON = 22;
const unsigned int DCF_INTERRUPT = 23;
DCF77 DCF = DCF77(DCF_PIN,DCF_INTERRUPT);

// EEPROM
unsigned int EEPROM_ADDR_HOUR = 0;
unsigned int EEPROM_ADDR_MIN = 1;
unsigned int EEPROM_ADDR_SONGNR = 2;

// 2 ISR for pushbuttons
void ISR_PB_Right() {
  mainEventFlags |= FLAG_INTERRUPT_PB_RIGHT;
}
void ISR_PB_Left() {
  mainEventFlags |= FLAG_INTERRUPT_PB_LEFT;
}

void setup() {
  /*setTime(17,45,00,28,07,15);  // sets the system time (NOT the Teensy RTC Clock)
  Teensy3Clock.set(now());// use the system time to update the Teensy's RTC*/
//  Serial.begin(115200);  
  delay(100);
//  Serial.println("Setup");
  // init DCF77 PON
  pinMode(DCF_PIN_PON, OUTPUT);
  digitalWrite(DCF_PIN_PON, HIGH); //DCF77 off
  // init 2 push buttons pins and interrupts
  pinMode(PB_RIGHT, INPUT_PULLUP);
  pinMode(PB_LEFT, INPUT_PULLUP);
  attachInterrupt(PB_RIGHT, ISR_PB_Right, FALLING);
  attachInterrupt(PB_LEFT, ISR_PB_Left, FALLING);
  // Init audio amplifier
  audioamp.begin();
  audioamp.enableChannel(false, false);  //turn off music: channel (right,left)
  audioamp.setAGCCompression(TPA2016_AGC_OFF); // turn off AGC
  audioamp.setReleaseControl(0); // turn off release to really turn off AGC
  audioamp.setGain(0); // set gain to 0
  // Audio connections require memory to work (see MemoryAndCpuUsage example)
  AudioMemory(5);
  // set the time library to use Teensy 3.1's RTC to keep time
  setSyncProvider(getTeensy3Time);
  if (timeStatus()!= timeSet) {
    //Serial.println("Unable to sync with the RTC");
  } else {
    //Serial.println("RTC has set the system time");
  }
  // init 7 segment backpack
  disp.begin(0x70); 
  disp.setBrightness(brightnessDay);  
  //init SD card  
  SPI.setMOSI(7);
  SPI.setSCK(14);
  SD.begin(10);
  /*if (!(SD.begin(10))) {
    // stop here, but print a message repetitively
    while (1) {
      Serial.println("Unable to access the SD card");
      delay(500);
    }
  }*/
  // get the number of files on disk
  filesondisk = filesOnDisk();
  //Serial.println(filesondisk);
  fileNr = 1;
  /*EEPROM.write(EEPROM_ADDR_HOUR, 6);
  EEPROM.write(EEPROM_ADDR_MIN, 00);
  EEPROM.write(EEPROM_ADDR_SONGNR, 1);*/
  //read alarm time from EEPROM
  alarmtimeHour = EEPROM.read(EEPROM_ADDR_HOUR);
  alarmtimeMin = EEPROM.read(EEPROM_ADDR_MIN);
  alarmtime = alarmtimeHour*100 + alarmtimeMin;  
  // init random generator: randomSeed(analogRead(A2));
  // for brightness change during CLOCK state
  knobRight.write(brightnessDay*4);
  positionRight = knobRight.read()/4;
  brightness = positionRight;
  rotaryTimer = 0;
}

void loop() {
  switch (state) {
    case CLOCK: {
      if (mainEventFlags & FLAG_SET_BRIGHTNESS) {
        rotaryTimer++;
        Serial.println(rotaryTimer);
        disp.print(brightness,DEC);
        disp.writeDigitNum(0,8);
        disp.writeDisplay();	
        if (rotaryTimer == 20) {
          mainEventFlags &= ~FLAG_SET_BRIGHTNESS; 
          rotaryTimer = 0;
	}
      }  
      else if (mainEventFlags & FLAG_ALARM_SET) displayClock(0x0A);
      else displayClock(0x02);
      newRight = knobRight.read();
      if (newRight/4 != positionRight) {
        mainEventFlags |= FLAG_SET_BRIGHTNESS;
        if (newRight < 0) {
          newRight = 0;  
          knobRight.write(0);
        }  
        if (newRight > 15*4) {
          newRight = 15*4;  
          knobRight.write(15*4);
        }
        positionRight = newRight/4;
        brightness = positionRight;
        disp.setBrightness(brightness);
        //Serial.println(positionRight);
        delayMicroseconds(10000); //10ms
        }
      // right button push: set alarm time
      if (mainEventFlags & FLAG_INTERRUPT_PB_RIGHT) {
        state = SET_ALARM;
        mainEventFlags &= ~FLAG_INTERRUPT_PB_RIGHT; //clear the Flag
        break;
      }
      // left button push: play music
      if (mainEventFlags & FLAG_INTERRUPT_PB_LEFT) {
        state = MUSIC;
        mainEventFlags &= ~FLAG_INTERRUPT_PB_LEFT; //clear the Flag
        break;
      }
      // test if alarm time reached
      if ((seg7time == alarmtime) && (second() == 0) &&
         (mainEventFlags & FLAG_ALARM_SET)) {
	state = ALARM;        
	break;
      }
      for (unsigned int i=0; i <= 10; i++) delayMicroseconds(10000); //100ms
      break;
    }
    case ALARM: {
      unsigned int manualvolume = false;
      unsigned int atimecounter = 0;
      unsigned int alarmgain = 0;
      rotaryTimer = 0;
      knobRight.write(0); // musicGain = 0
      positionRight = knobRight.read()/4;
      musicGain = positionRight;
      audioamp.setGain(0); 
      audioamp.enableChannel(true, false);
      //random file to play
      fileNr = EEPROM.read(EEPROM_ADDR_SONGNR);
      sprintf(fileName, "%01d.WAV", fileNr);
      playWav1.play(fileName);
      delayMicroseconds(5000);  // A brief delay for the library read WAV info
      while (!((mainEventFlags & FLAG_INTERRUPT_PB_RIGHT) ||
             (mainEventFlags & FLAG_INTERRUPT_PB_LEFT))) {
        if (mainEventFlags & FLAG_SET_GAIN) {
          rotaryTimer++;          
          disp.print(musicGain,DEC);
          disp.writeDigitNum(0,6);
          disp.writeDisplay();	
          if (rotaryTimer == 100) {
            mainEventFlags &= ~FLAG_SET_GAIN; 
            rotaryTimer = 0;
	  }
        }  
        else {
          displayClock(0x0A);
          rotaryTimer = 0;
        }
        delayMicroseconds(10000); //10ms
        atimecounter++;
        newRight = knobRight.read();
        if (newRight/4 != positionRight) {
          mainEventFlags |= FLAG_SET_GAIN;
          manualvolume = true;
          if (newRight < 0) {
            newRight = 0;  
            knobRight.write(0);
          }
          if (newRight > 30*4) {
            newRight = 30*4;  
            knobRight.write(30*4);
          }          
          positionRight = newRight/4;
          musicGain = positionRight;
          audioamp.setGain(musicGain); 
          delayMicroseconds(10000); //10ms
        }
        if ((atimecounter%300 == 0) && (!manualvolume)) { //300 for 3 seconds
          if (alarmgain != 30) alarmgain +=1; 
          //Serial.println(alarmgain);
          audioamp.setGain(alarmgain); 
        }  
        if (!playWav1.isPlaying()) {
          fileNr++;
          if (fileNr > filesondisk) fileNr = 1;
          sprintf(fileName, "%01d.WAV", fileNr);
          //Serial.println(fileName);
          playWav1.play(fileName);
          delayMicroseconds(5000);  // A brief delay for the library read WAV info
        }  
      }  
      audioamp.setGain(0); 
      audioamp.enableChannel(false, false);	 
      playWav1.stop(); 
      manualvolume = false;
      fileNr = 1;      
      mainEventFlags &= ~FLAG_INTERRUPT_PB_RIGHT; //clear the Flags
      mainEventFlags &= ~FLAG_INTERRUPT_PB_LEFT; 
      mainEventFlags &= ~FLAG_ALARM_SET; //clear the Flag
      state = CLOCK;
      break;
    }
    case SET_ALARM: {
      knobLeft.write(alarmtimeHour*4);
      knobRight.write(alarmtimeMin*4);
      positionLeft = alarmtimeHour;
      positionRight = alarmtimeMin;
      while (!((mainEventFlags & FLAG_INTERRUPT_PB_RIGHT) ||
             (mainEventFlags & FLAG_INTERRUPT_PB_LEFT))) {
        disp.print(alarmtime,DEC);
        if (alarmtime <= 59) disp.writeDigitNum(1,0);
        if (alarmtime <= 9) disp.writeDigitNum(3,0);
        disp.writeDigitRaw(2,0x08); // 0x08 left colon lower dot no center colon
        disp.writeDisplay();        
        newLeft = knobLeft.read();
        if (newLeft/4 != positionLeft) {
          if (newLeft < 0) {  // clip to min and max
            newLeft = 0;  
            knobLeft.write(0);
          }
          if (newLeft > 23*4) {
            newLeft = 23*4;  
            knobLeft.write(23*4);
          }
          positionLeft = newLeft/4;
        }
        newRight = knobRight.read();
        if (newRight/4 != positionRight) {
          if (newRight < 0) {
            newRight = 0;  
            knobRight.write(0);
          }
          if (newRight > 59*4) {
            newRight = 59*4;  
            knobRight.write(59*4);
          }
          positionRight = newRight/4;
        }
        alarmtimeHour = positionLeft;
        alarmtimeMin = positionRight;
        alarmtime = alarmtimeHour*100+alarmtimeMin;
      } 
      if (mainEventFlags & FLAG_INTERRUPT_PB_RIGHT) {
        mainEventFlags |= FLAG_ALARM_SET;
        EEPROM.write(EEPROM_ADDR_HOUR,alarmtimeHour);
        EEPROM.write(EEPROM_ADDR_MIN,alarmtimeMin);
      }  
      if (mainEventFlags & FLAG_INTERRUPT_PB_LEFT) {
         mainEventFlags &= ~FLAG_ALARM_SET;
      }
      for (unsigned int i=0; i <= 20; i++) delayMicroseconds(10000); // 200ms
      mainEventFlags &= ~FLAG_INTERRUPT_PB_RIGHT; //clear the Flags
      mainEventFlags &= ~FLAG_INTERRUPT_PB_LEFT; 
      state = CLOCK;
      break;
    }  
    case MUSIC: {
      //Serial.println("Music ");
      rotaryTimer = 0;
      positionLeft = knobLeft.read()/4;
      fileNr = positionLeft;
      positionRight = knobRight.read()/4;
      musicGain = positionRight;
      //Serial.println(positionRight);
      audioamp.enableChannel(true, false);
      audioamp.setGain(musicGain);
      //random file to play:  fileNr = random(1,filesondisk+1);
      sprintf(fileName, "%01d.WAV", fileNr);
      //Serial.println(fileName);
      playWav1.play(fileName);
      delayMicroseconds(5000); //5ms A brief delay for the library read WAV info
      while (!((mainEventFlags & FLAG_INTERRUPT_PB_RIGHT) ||
            (mainEventFlags & FLAG_INTERRUPT_PB_LEFT))) {        
        if (mainEventFlags & FLAG_SET_SONG) {
	  rotaryTimer++;
          //Serial.println(rotaryTimer);
          disp.print(fileNr,DEC);
          disp.writeDigitNum(0,5);
          disp.writeDisplay();	
          if (rotaryTimer == 1000) {
            mainEventFlags &= ~FLAG_SET_SONG; 
            rotaryTimer = 0;
	  }
        }  
	else if (mainEventFlags & FLAG_SET_GAIN) {
          rotaryTimer++;
          //Serial.println(rotaryTimer);
          disp.print(musicGain,DEC);
          disp.writeDigitNum(0,6);
          disp.writeDisplay();	
          if (rotaryTimer == 1000) {
            mainEventFlags &= ~FLAG_SET_GAIN; 
            rotaryTimer = 0;
	  }
        }  
        else {
          displayClock(0x06);
          rotaryTimer = 0;
        }
        newLeft = knobLeft.read();
        if (newLeft/4 != positionLeft) {
	  mainEventFlags |= FLAG_SET_SONG;
          if (newLeft < 4) {  // clip to min and max
            newLeft = 4;  
            knobLeft.write(4);
          }
          if (newLeft > filesondisk*4) {
            newLeft = filesondisk*4;  
            knobLeft.write(filesondisk*4);
          }
          positionLeft = newLeft/4;  
	  fileNr = positionLeft;
          sprintf(fileName, "%01d.WAV", fileNr);
          //Serial.println(fileName);          
          playWav1.stop(); 
          playWav1.play(fileName);
          delayMicroseconds(10000); //10ms
        }
        newRight = knobRight.read();
        if (newRight/4 != positionRight) {
          mainEventFlags |= FLAG_SET_GAIN;
          if (newRight < 0) {
            newRight = 0;  
            knobRight.write(0);
          }  
          if (newRight > 30*4) {
            newRight = 30*4;  
            knobRight.write(30*4);
          }
          positionRight = newRight/4;
          musicGain = positionRight;
          audioamp.setGain(musicGain); 
          //Serial.println(positionRight);
	  delayMicroseconds(10000); //10ms
        }
        if (!playWav1.isPlaying()) {
          fileNr++;
          if (fileNr > filesondisk) fileNr = 1;
          sprintf(fileName, "%01d.WAV", fileNr);
          playWav1.play(fileName);
          delayMicroseconds(5000); //5ms
        }  
      }
      if (mainEventFlags & FLAG_INTERRUPT_PB_RIGHT) {
        EEPROM.write(EEPROM_ADDR_SONGNR,fileNr);        
      }  
      delayMicroseconds(10000); //10ms
      audioamp.setGain(0); 
      audioamp.enableChannel(false, false);	 
      playWav1.stop(); 
      mainEventFlags &= ~FLAG_INTERRUPT_PB_RIGHT; //clear the Flags
      mainEventFlags &= ~FLAG_INTERRUPT_PB_LEFT; 
      state = CLOCK;
      break;
    }
    case DCF_77: { 
      mainEventFlags &= ~FLAG_DCF77_ERROR;
      digitalWrite(DCF_PIN_PON, LOW); //DCF77 on
      for (unsigned int i=0; i <= 100; i++) delayMicroseconds(10000); // 1s
      DCF.Start();
      for (unsigned int i=0; i <= 300; i++) delayMicroseconds(10000); // 3s
      for (unsigned int i=0; i <= 600; i++) { // 10 minutes to get DCF-time
        if (mainEventFlags & FLAG_ALARM_SET) displayClock(0x0A);
        else displayClock(0x02);
	for (unsigned int i=0; i <= 100; i++) delayMicroseconds(10000); // 1s
	time_t DCFtime = DCF.getTime();
	if (DCFtime==0) {
          mainEventFlags |= FLAG_DCF77_ERROR; 
        }  
        else {
          Teensy3Clock.set(DCFtime);
          mainEventFlags &= ~FLAG_DCF77_ERROR; 
          break;
        }  
      }
      DCF.Stop();
      state = CLOCK;
      break;
    }    
  }
}

time_t getTeensy3Time() {
  return Teensy3Clock.get();
}

//check the number of files on disk
unsigned int filesOnDisk() {
char fileName[13];
unsigned int fileNr = 1;
while (1) {
    sprintf(fileName, "%01d.WAV", fileNr);    
    if (SD.exists(fileName) == false) break;
    fileNr++;
   }
  return fileNr-1;
}

// display the clock
// dotState = 0x02: only center colon, normal clock display
//            0x0A: 0x02 center colon, 0x08 left colon lower set alarm
//            0x06: 0x02 center colon, 0x04 left colon upper music
void displayClock(int dotState) {
  getTeensy3Time();  
  seg7time = hour()*100+minute();
  disp.print(seg7time,DEC);
  if (seg7time <= 59) disp.writeDigitNum(1,0);
  if (seg7time <= 9) disp.writeDigitNum(3,0);
  if (mainEventFlags & FLAG_DCF77_ERROR) dotState = dotState | 0x10;
  disp.writeDigitRaw(2,dotState);
  if (seg7time == brightnessDayTime) disp.setBrightness(brightnessDay);  
  if (seg7time == brightnessNightTime) disp.setBrightness(brightnessNight);  
  if (seg7time == DCF77Time) state = DCF_77;
  disp.writeDisplay();
}



// only for debugging:
/*void printDigits(int digits) {
  // utility function for digital clock display: prints preceding colon and leading 0
  Serial.print(":");
  if(digits < 10)
    Serial.print('0');
  Serial.print(digits);
}

void serialClockDisplay() {
  // digital clock display of the time
  Serial.print(hour());
  printDigits(minute());
  printDigits(second());
  Serial.print(" ");
  Serial.print(day());
  Serial.print(" ");
  Serial.print(month());
  Serial.print(" ");
  Serial.print(year()); 
  Serial.println(); 
}*/
  

//  writeDigitRaw(2,0x02);  //  0x02 - center colon (Doppelpunkt)
                            //  0x04 - left colon upper dot
                            //  0x08 - left colon lower dot
                            //  0x10 - upper right decimal point
  

        //disp.writeDigitNum(4,alarmtimeMin%10);
        //disp.writeDigitNum(3,alarmtimeMin/10);
        //disp.writeDigitNum(1,alarmtimeHour%10);
        //disp.writeDigitNum(0,alarmtimeHour/10);

